"""Bits and pieces."""

import inspect
import os
import random
import socket
import sys
import time
from pathlib import Path
from typing import Any, Callable, Generator, Generic, Iterable, NamedTuple, TypeVar

DEBUG = os.getenv("DEBUG") not in ["0", "false", "False", None]

A = TypeVar("A")
B = TypeVar("B")
T = TypeVar("T")


class Retry:
    """Retry variant."""


class Ok(NamedTuple, Generic[T]):
    """Success variant."""

    result: T


type Result[T] = Ok[T] | RuntimeError | Retry


def flatmap(f: Callable[[A], Iterable[B]], xs: Iterable[A]) -> Iterable[B]:
    """Map f over an iterable and flatten the result set."""
    return (y for x in xs for y in f(x))


def log(
    msg: str, /, stack_offset: int = 1, file: Any = sys.stdout, skip_prefix: bool = False
) -> None:
    if skip_prefix:
        prefix = ""
    else:
        frame = inspect.stack()[stack_offset]
        module = inspect.getmodule(frame[0])
        if module is not None and getattr(module, "__file__") is not None:
            module_name = Path(getattr(module, "__file__")).with_suffix("").name
        else:
            module_name = "?"
        prefix = f"[{module_name}:{frame.function}]: "
    print(f"{prefix}{msg}", file=file)


def debug(msg: str) -> None:
    if DEBUG:
        log(msg, stack_offset=2, file=sys.stderr)


def retries[
    **P
](num_retries: int, sleep_time: float) -> Callable[
    [Callable[P, Result | Retry]], Callable[P, Result]
]:
    """ "just put a retry loop around it" """

    def wrapper(fn: Callable[P, Result | Retry]) -> Callable[P, Result]:
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> Result:
            for _ in range(0, num_retries):
                match fn(*args, **kwargs):
                    case Ok(x):
                        return x
                    case Retry():
                        time.sleep(sleep_time)
                        continue
                    case RuntimeError() as e:
                        return e
            return RuntimeError(f"retries exceeded in '{fn.__name__}'")

        return wrapped

    return wrapper


def find_free_port(min_port: int = 8100, max_port: int = 8500) -> int | None:
    num_tries = 0
    random_port = random.randint(min_port, max_port)
    while not is_port_open(random_port):
        if num_tries > 100:
            return None
        random_port = random.randint(min_port, max_port)
        num_tries += 1
    return random_port


def is_port_open(port: int) -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    res = sock.connect_ex(("127.0.0.1", port))
    sock.close()
    return res != 0


def remove_consecutive_falsy(xs: Iterable[T]) -> Generator[T, None, None]:
    yielded_empty = False
    for line in xs:
        if line:
            yielded_empty = False
            yield line
        elif not yielded_empty:
            yield line
            yielded_empty = True



"""GitHub actor."""

import re
from contextlib import suppress
from subprocess import CalledProcessError, check_output
from threading import Event
from typing import Callable, List, NamedTuple, Set, cast

from github import Repository
from github.WorkflowJob import WorkflowJob
from github.WorkflowRun import WorkflowRun
from pykka import ActorRef, ThreadingActor

from octotail.cli import Opts
from octotail.utils import Ok, Result, Retry, debug, log, retries

VALID_STATI = ["queued", "in_progress", "requested", "waiting", "action_required"]


class WorkflowDone(NamedTuple):
    """Workflow done message."""

    conclusion: str


class JobDone(NamedTuple):
    """Job done message."""

    job_id: int
    job_name: str
    conclusion: str


class RunWatcher(ThreadingActor):
    """Watches for changes in a GitHub Actions run."""

    wf_run: WorkflowRun
    mgr: ActorRef
    stop: Event

    _new_jobs: Set[int] = set()
    _concluded_jobs: Set[int] = set()

    def __init__(self, mgr: ActorRef, wf_run: WorkflowRun):
        super().__init__()
        self.wf_run = wf_run
        self.mgr = mgr
        self.stop = mgr.proxy().stop_event.get()

    def watch(self) -> None:
        while not self.stop.is_set():
            self.wf_run.update()

            with suppress(Exception):
                for job in self.wf_run.jobs():
                    if job.conclusion and not job.id in self._concluded_jobs:
                        if not self._tell(JobDone(job.id, job.name, job.conclusion)):
                            return
                        self._concluded_jobs.add(job.id)
                        continue
                    if not job.id in self._new_jobs.union(self._concluded_jobs):
                        if not self._tell(job):
                            return
                        self._new_jobs.add(job.id)

            if self.wf_run.conclusion:
                self._tell(WorkflowDone(self.wf_run.conclusion))
                return

            self.stop.wait(2)
        debug("exiting")

    def _tell(self, what: JobDone | WorkflowDone | WorkflowJob) -> bool:
        if self.mgr.is_alive():
            self.mgr.tell(what)
            return True
        return False


@retries(10, 0.5)
def get_active_run(repo: Repository, opts: Opts) -> Result[WorkflowRun] | Retry:
    with suppress(Exception):
        runs = repo.get_workflow_runs(head_sha=opts.commit_sha)
        if runs.totalCount == 0:
            return Retry()

        filters: List[Callable[[WorkflowRun], bool]] = [lambda wf: wf.status in VALID_STATI]
        if opts.workflow_name:
            filters.append(lambda wf: wf.name == cast(str, opts.workflow_name))
        if opts.ref_name:
            filters.append(lambda wf: cast(str, opts.ref_name).endswith(wf.head_branch))

        _runs = [r for r in runs if all(f(r) for f in filters)]
        if not _runs:
            return Retry()

        if len(_runs) > 1:
            log(f"found multiple active runs for commit '{opts.commit_sha}'")
            for run in _runs:
                log(f"\n\t{run.html_url}", skip_prefix=True)
            log("", skip_prefix=True)
            log("try narrowing down by workflow name (--workflow) or ref name (--ref-name)")
            return RuntimeError("cannot disambiguate")

        return Ok(runs[0])

    return Retry()


def guess_repo() -> str | None:
    try:
        remotes = check_output(["git", "remote"]).decode().strip().splitlines()
    except CalledProcessError as e:
        log(f"fatal: couldn't list git remotes: {e}")
        return None
    repos = list(filter(None, map(_guess_repo, remotes)))
    if len(repos) > 1:
        log("fatal: found multiple remotes pointing to GitHub")
        return None
    return repos[0] if repos else None


def _guess_repo(remote: str) -> str | None:
    with suppress(Exception):
        origin = check_output(f"git remote get-url {remote}".split()).decode().strip()
        if (match := re.search(r"^git@github.com:([^.]+).git$", origin)) is not None:
            return match.group(1)
    return None



#!/usr/bin/env -S /bin/sh -c 'exec "$(dirname $(readlink -f "$0"))/../.venv/bin/python3" "$0" "$@"'

"""Having your cake and eating it too."""
import dataclasses
import multiprocessing as mp
import sys
from pathlib import Path
from threading import Event
from typing import Dict, Union

from github import Auth, Github
from github.WorkflowJob import WorkflowJob
from github.WorkflowRun import WorkflowRun
from pykka import ActorRegistry, ThreadingActor
from xdg.BaseDirectory import xdg_cache_home

from octotail.browser import BrowseRequest, BrowserWatcher, CloseRequest, ExitRequest, VisitRequest
from octotail.cli import Opts, entrypoint
from octotail.fmt import Formatter
from octotail.gh import JobDone, RunWatcher, WorkflowDone, get_active_run, guess_repo
from octotail.mitm import ProxyWatcher, WsSub
from octotail.streamer import OutputItem, run_streamer
from octotail.utils import debug, find_free_port, log

COOKIE_JAR = Path(xdg_cache_home) / "octotail" / "gh-cookies.json"

type MgrMessage = Union[WorkflowJob, WsSub, JobDone, WorkflowDone]


class Manager(ThreadingActor):
    """I'm the Baahwss."""

    browse_queue: mp.Queue
    output_queue: mp.JoinableQueue
    stop_event: Event

    streamers: Dict[int, mp.Process]
    job_map: Dict[int, str]

    def __init__(self, browse_queue: mp.Queue, output_queue: mp.JoinableQueue, stop: Event):
        super().__init__()
        self.browse_queue = browse_queue
        self.output_queue = output_queue
        self.stop_event = stop

        self.streamers = {}
        self.job_map = {}

    def on_receive(self, msg: MgrMessage) -> None:
        debug(f"manager got message: {msg!r}")

        match msg:
            case WorkflowJob() as job:
                self.browse_queue.put_nowait(VisitRequest(job.html_url, job.id))
                self.job_map[job.id] = job.name

            case WsSub() as ws_sub:
                self.browse_queue.put_nowait(CloseRequest(ws_sub.job_id))
                if ws_sub.job_id in self.job_map:
                    ws_sub = dataclasses.replace(ws_sub, job_name=self.job_map[ws_sub.job_id])
                self._replace_streamer(ws_sub.job_id, run_streamer(ws_sub, self.output_queue))

            case JobDone() as job:
                self.output_queue.put(OutputItem(job.job_name, [f"##[conclusion]{job.conclusion}"]))
                self._terminate_streamer(job.job_id)

            case WorkflowDone() as wf_done:
                self.output_queue.put(
                    OutputItem("workflow", [f"##[conclusion]{wf_done.conclusion}"])
                )
                self.stop()

    def on_stop(self) -> None:
        self.stop_event.set()
        self.browse_queue.put_nowait(ExitRequest())
        self.output_queue.put_nowait(None)
        for p in self.streamers.values():
            p.terminate()
        debug("manager exiting")

    def _terminate_streamer(self, job_id: int) -> None:
        if job_id in self.streamers:
            self.streamers[job_id].terminate()
            del self.streamers[job_id]

    def _replace_streamer(self, job_id: int, streamer: mp.Process) -> None:
        self._terminate_streamer(job_id)
        self.streamers[job_id] = streamer


@entrypoint
def _main(opts: Opts) -> None:
    _stop = Event()

    if (repo_id := (opts.repo or guess_repo())) is None:
        log("could not establish repo")
        sys.exit(1)

    gh_client = Github(auth=Auth.Token(opts.gh_pat))
    wf_run = get_active_run(gh_client.get_repo(repo_id), opts)
    if not isinstance(wf_run, WorkflowRun):
        log(f"could not get a matching workflow run: {wf_run}")
        sys.exit(1)

    # find a free port
    if opts.port is None:
        if (port := find_free_port()) is not None:
            opts = dataclasses.replace(opts, port=port)
        else:
            log("giving up finding a free port in the 8100 - 8500 range")
            sys.exit(1)

    debug(f"starting on port {opts.port}")

    # pylint: disable=E1136
    browser_inbox: mp.Queue[BrowseRequest] = mp.Queue()
    output_queue: mp.JoinableQueue[OutputItem | None] = mp.JoinableQueue()
    manager = Manager.start(browser_inbox, output_queue, _stop)

    run_watcher = RunWatcher.start(manager, wf_run)
    browser_watcher = BrowserWatcher.start(manager, opts, browser_inbox)
    proxy_watcher = ProxyWatcher.start(manager, opts.port)
    formatter = Formatter.start(output_queue)

    try:
        run_watcher.proxy().watch().join(
            browser_watcher.proxy().watch(),
            proxy_watcher.proxy().watch(),
            formatter.proxy().print_lines(),
        ).get()
    except KeyboardInterrupt:
        _stop.set()

    ActorRegistry.stop_all()


if __name__ == "__main__":
    _main()



"""Mitmproxy actor."""

import base64
import copy
import json
import multiprocessing
import sys
from argparse import Namespace
from contextlib import suppress
from dataclasses import dataclass
from queue import Empty
from threading import Event
from typing import List

from mitmproxy.tools.main import mitmdump
from pykka import ActorRef, ThreadingActor

from octotail.utils import Ok, Result, Retry, debug, is_port_open, retries

MARKERS = Namespace(
    ws_header="WebSocket text message",
    ws_host="alive.github.com",
    ws_action='"subscribe":',
)


@dataclass(frozen=True)
class WsSub:
    """Represents a websocket subscription."""

    url: str
    subs: str
    job_id: int
    job_name: str | None = None


class ProxyWatcher(ThreadingActor):
    """Watches for websocket subscriptions done through the mitmproxy."""

    mgr: ActorRef
    port: int
    _stop_event: Event
    _proxy_ps: multiprocessing.Process
    _q: multiprocessing.Queue

    def __init__(self, mgr: ActorRef, port: int):
        super().__init__()
        self.mgr = mgr
        self.port = port
        self._stop_event = mgr.proxy().stop_event.get()

    def on_start(self) -> None:
        self._q = multiprocessing.Queue()
        self._proxy_ps = multiprocessing.Process(
            target=_mitmdump_wrapper, args=(self._q, self.port)
        )
        self._proxy_ps.start()

    def on_stop(self) -> None:
        self._proxy_ps.terminate()
        self._proxy_ps.join()

    def watch(self) -> None:
        if not _check_liveness(self._proxy_ps, self.port):
            self.mgr.tell("fatal: proxy didn't go live")
            self.mgr.stop()

        old_line, buffer = "-", []
        old_buffer: List[str] = []
        while not self._stop_event.is_set():
            with suppress(Empty):
                line = self._q.get(timeout=1).strip()
                if line:
                    buffer.append(line)
                elif old_line == "" and buffer:
                    self._process_buffer("".join(buffer), "".join(old_buffer))
                    old_buffer = copy.deepcopy(buffer)
                    buffer = []
                else:
                    buffer.append(line)
                old_line = line
        debug("exiting")

    def _process_buffer(self, buffer: str, old_buffer: str) -> None:
        if (
            MARKERS.ws_header in old_buffer
            and MARKERS.ws_host in old_buffer
            and MARKERS.ws_action in buffer
        ):
            url = old_buffer[old_buffer.index(MARKERS.ws_host) :]
            if (job_id := _extract_job_id(buffer)) is not None:
                self.mgr.tell(WsSub(url=url, subs=buffer, job_id=job_id))


def _extract_job_id(buffer: str) -> int | None:
    with suppress(Exception):
        dec = (base64.b64decode(k) for k in json.loads(buffer)["subscribe"].keys())
        items = (str(json.loads(d[: d.index(b"}") + 1].decode())["c"]) for d in dec)
        good = next(item for item in items if item.startswith("check_runs"))
        return int(good.split(":")[1])
    return None


def _mitmdump_wrapper(q: multiprocessing.Queue, port: int) -> None:
    sys.argv = f"mitmdump --flow-detail=4 --no-rawtcp -p {port}".split()
    setattr(sys.stdout, "isatty", lambda: False)
    setattr(sys.stdout, "write", q.put)
    mitmdump()


@retries(50, 0.2)
def _check_liveness(proxy_ps: multiprocessing.Process, port: int) -> Result[bool] | Retry:
    if not is_port_open(port):
        return Ok(True)
    if not proxy_ps.is_alive():
        return Ok(False)
    return Retry()



""" Options class + some magic so we can define our args in one place only. """

import re
from contextvars import ContextVar
from dataclasses import dataclass
from functools import wraps
from typing import Annotated, Callable

from typer import Argument, BadParameter, Option, Typer

_REPO_HELP = "\n".join(
    [
        "Use this GitHub repo to look for workflow runs.",
        "If unspecified, will look for a remote matching 'git@github.com:<user>/<repo>.git'",
        "in the current directory.",
        "\nExamples: 'user/repo' OR 'org_name/repo'",
    ]
)


def _sha_callback(value: str) -> str:
    if len(value) != 40:
        raise BadParameter("need a full 40 character long commit sha")
    if value == 40 * "0":
        raise BadParameter("refusing to work with the all-zero commit sha")
    return value


def _repo_callback(value: str | None) -> str | None:
    pattern = r"^[a-zA-Z0-9_-]{1,100}/[a-zA-Z0-9_-]{1,100}$"
    if value is not None and re.match(pattern, value) is None:
        raise BadParameter(f"invalid format for repo: {value}")
    return value


@dataclass(frozen=True)
class Opts:
    """
    Look for an active workflow run for the given <COMMIT_SHA> (and optionally
    --workflow and/or --ref-name) and attempt to tail its logs.

    NOTE: the <COMMIT_SHA> has to be of the full 40 characters length.
    """

    commit_sha: Annotated[
        str,
        Argument(callback=_sha_callback, help="Full commit SHA that triggered the workflow."),
    ]
    gh_pat: Annotated[
        str,
        Option(
            envvar="_GH_PAT",
            help="GitHub personal access token. (for API auth)",
            show_default=False,
        ),
    ]
    gh_user: Annotated[
        str,
        Option(envvar="_GH_USER", help="GitHub username. (for web auth)", show_default=False),
    ]
    gh_pass: Annotated[
        str,
        Option(envvar="_GH_PASS", help="GitHub password. (for web auth)", show_default=False),
    ]
    gh_otp: Annotated[str | None, Option(envvar="_GH_OTP", help="GitHub OTP. (for web auth)")] = (
        None
    )
    workflow_name: Annotated[
        str | None,
        Option(
            "-w",
            "--workflow",
            help="Only consider workflows with this name.",
            show_default=False,
        ),
    ] = None
    ref_name: Annotated[
        str | None,
        Option(
            "-r",
            "--ref-name",
            help="Only consider workflows triggered by this ref.\n\nExample: 'refs/heads/main'",
            show_default=False,
        ),
    ] = None
    repo: Annotated[
        str | None,
        Option(
            "-R",
            "--repo",
            help=_REPO_HELP,
            show_default=False,
            callback=_repo_callback,
        ),
    ] = None
    headless: Annotated[bool, Option(envvar="_HEADLESS")] = True
    port: Annotated[
        int | None, Option(envvar="_PORT", show_default="random in range 8100-8500")
    ] = None

    def __post_init__(self) -> None:
        _post_init.get()(self)


def _noop(_: Opts) -> None:
    pass


_post_init: ContextVar[Callable[[Opts], None]] = ContextVar("post_init", default=_noop)


def entrypoint(main_fn: Callable[[Opts], None]) -> Callable:
    def wrapped(opts: Opts) -> None:
        _post_init.set(_noop)
        main_fn(opts)

    @wraps(main_fn)
    def wrapper() -> None:
        app = Typer(add_completion=False, rich_markup_mode="rich")
        app.command(no_args_is_help=True)(Opts)
        _post_init.set(wrapped)
        app()

    return wrapper



"""Streamer actor."""

import asyncio as aio
import json
import multiprocessing as mp
from contextlib import suppress
from typing import List, NamedTuple

import websockets.client

from octotail.browser import RANDOM_UA
from octotail.mitm import WsSub
from octotail.utils import log

WS_HEADERS = {
    "User-Agent": RANDOM_UA,
    "Origin": "https://github.com",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "websocket",
    "Sec-Fetch-Site": "same-site",
    "Pragma": "no-cache",
    "Cache-Control": "no-cache",
}


class OutputItem(NamedTuple):
    """Holds an output item."""

    job_name: str
    lines: List[str]


def run_streamer(ws_sub: WsSub, queue: mp.Queue) -> mp.Process:
    process = mp.Process(target=_streamer, args=(ws_sub, queue))
    process.start()
    return process


def _streamer(ws_sub: WsSub, queue: mp.Queue) -> None:
    loop = aio.new_event_loop()
    aio.set_event_loop(loop)
    try:
        loop.run_until_complete(_stream_it(ws_sub, queue))
    except KeyboardInterrupt:
        loop.close()


async def _stream_it(ws_sub: WsSub, queue: mp.Queue) -> None:
    ws_url = "wss://" + ws_sub.url.removeprefix("https://")
    job_name = ws_sub.job_name or "unknown"

    async for websocket in websockets.client.connect(ws_url, extra_headers=WS_HEADERS):
        try:
            await websocket.send(ws_sub.subs)
            async for msg in websocket:
                lines = _extract_lines(msg)
                if lines:
                    queue.put(OutputItem(job_name, lines))
        except aio.CancelledError:
            log("cancelled")
            if websocket:
                await websocket.close()
            return


def _extract_lines(msg: str) -> List[str]:
    with suppress(Exception):
        return [l["line"] for l in json.loads(msg)["data"]["data"]["lines"]]
    return []
